STARS = {
    "Posting": "11.4k",
    "Memray": "14.9k",
    "Toolong": "3.9k",
    "Dolphie": "1.1k",
    "Harlequin": "5.8k",
    "Elia": "2.4k",
    "Trogon": "2.8k",
    "TFTUI - The Terraform textual UI": "1.3k",
    "RecoverPy": "1.7k",
    "Frogmouth": "3.1k",
    "oterm": "2.3k",
    "logmerger": "250",
    "doit": "2.8k",
}



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from textual.app import RenderResult

from textual.visual import Visual, VisualType, visualize
from textual.widget import Widget


class Static(Widget, inherit_bindings=False):
    """A widget to display simple static content, or use as a base class for more complex widgets.

    Args:
        content: A Content object, Rich renderable, or string containing console markup.
        expand: Expand content if required to fill container.
        shrink: Shrink content if required to fill container.
        markup: True if markup should be parsed and rendered.
        name: Name of widget.
        id: ID of Widget.
        classes: Space separated list of class names.
        disabled: Whether the static is disabled or not.
    """

    DEFAULT_CSS = """
    Static {
        height: auto;
    }
    """

    def __init__(
        self,
        content: VisualType = "",
        *,
        expand: bool = False,
        shrink: bool = False,
        markup: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name, id=id, classes=classes, disabled=disabled, markup=markup
        )
        self.set_reactive(Widget.expand, expand)
        self.set_reactive(Widget.shrink, shrink)
        self.__content = content
        self.__visual: Visual | None = None

    @property
    def visual(self) -> Visual:
        """The visual to be displayed.

        Note that the visual is what is ultimately rendered in the widget, but may not be the
        same object set with the `update` method  or `content` property. For instance, if you
        update with a string, then the visual will be a [Content][textual.content.Content] instance.

        """
        if self.__visual is None:
            self.__visual = visualize(self, self.__content, markup=self._render_markup)
        return self.__visual

    @property
    def content(self) -> VisualType:
        """The original content set in the constructor."""
        return self.__content

    @content.setter
    def content(self, content: VisualType) -> None:
        self.__content = content
        self.__visual = visualize(self, content, markup=self._render_markup)
        self.clear_cached_dimensions()
        self.refresh(layout=True)

    def render(self) -> RenderResult:
        """Get a rich renderable for the widget's content.

        Returns:
            A rich renderable.
        """
        return self.visual

    def update(self, content: VisualType = "", *, layout: bool = True) -> None:
        """Update the widget's content area with a string, a Visual (such as [Content][textual.content.Content]), or a [Rich renderable](https://rich.readthedocs.io/en/latest/protocol.html).

        Args:
            content: New content.
            layout: Also perform a layout operation (set to `False` if you are certain the size won't change).
        """

        self.__content = content
        self.__visual = visualize(self, content, markup=self._render_markup)
        self.refresh(layout=layout)



from __future__ import annotations

from textual.widgets import Static


class Tooltip(Static, inherit_css=False):
    DEFAULT_CSS = """
    Tooltip {
        layer: _tooltips;
        margin: 1 0;
        padding: 1 2;
        background: $panel;
        width: auto;
        height: auto;
        constrain: inside inflect;
        max-width: 40;
        display: none;
        offset-x: -50%;
        &:ansi {
            background: $ansi-background;
        }
    }
    """
    DEFAULT_CLASSES = "-textual-system"



from __future__ import annotations

from itertools import zip_longest
from typing import TYPE_CHECKING

from textual.geometry import NULL_OFFSET, Region, Size
from textual.layout import ArrangeResult, Layout, WidgetPlacement

if TYPE_CHECKING:
    from textual.widget import Widget


class StreamLayout(Layout):
    """A cut down version of the vertical layout.

    The stream layout is faster, but has a few limitations compared to the vertical layout.

    - All widgets are the full width (as if their widget is `1fr`).
    - All widgets have an effective height of `auto`.
    - `max-height` is supported, but only if it is a units value, all other extrema rules are ignored.
    - No absolute positioning.
    - No overlay: screen.
    - Layers are ignored.
    - Non TCSS styles are ignored.

    The primary use of `layout: stream` is for a long list of widgets in a scrolling container, such as
    what you might expect from a LLM chat-bot. The speed improvement will only be significant with a lot of
    child widgets, so stick to vertical layouts unless you see any slowdown.

    """

    name = "stream"

    def __init__(self) -> None:
        self._cached_placements: list[WidgetPlacement] | None = None
        self._cached_width = 0
        super().__init__()

    def arrange(
        self, parent: Widget, children: list[Widget], size: Size, greedy: bool = True
    ) -> ArrangeResult:
        parent.pre_layout(self)
        if not children:
            return []
        viewport = parent.app.viewport_size

        if size.width != self._cached_width:
            self._cached_placements = None
        previous_results = self._cached_placements or []

        layout_widgets = parent.screen._layout_widgets.get(parent, [])

        _Region = Region
        _WidgetPlacement = WidgetPlacement

        placements: list[WidgetPlacement] = []
        width = size.width
        first_child_styles = children[0].styles
        y = 0
        previous_margin = first_child_styles.margin.top
        null_offset = NULL_OFFSET

        pre_populate = bool(previous_results and layout_widgets)
        for widget, placement in zip_longest(children, previous_results):
            if pre_populate and placement is not None and widget is placement.widget:
                if widget in layout_widgets:
                    pre_populate = False
                else:
                    placements.append(placement)
                    y = placement.region.bottom
                    styles = widget.styles._base_styles
                    previous_margin = styles.margin.bottom
                    continue
            if widget is None:
                break

            styles = widget.styles._base_styles
            margin = styles.margin
            gutter_width, gutter_height = styles.gutter.totals
            top, right, bottom, left = margin
            y += top if top > previous_margin else previous_margin
            previous_margin = bottom
            height = (
                widget.get_content_height(size, viewport, width - gutter_width)
                + gutter_height
            )
            if (max_height := styles.max_height) is not None and max_height.is_cells:
                height = (
                    height
                    if height < (max_height_value := int(max_height.value))
                    else max_height_value
                )
            if (min_height := styles.min_height) is not None and min_height.is_cells:
                height = (
                    height
                    if height > (min_height_value := int(min_height.value))
                    else min_height_value
                )
            placements.append(
                _WidgetPlacement(
                    _Region(left, y, width - (left + right), height),
                    null_offset,
                    margin,
                    widget,
                    0,
                    False,
                    False,
                    False,
                )
            )
            y += height

        self._cached_width = size.width
        self._cached_placements = placements
        return placements

    def get_content_width(self, widget: Widget, container: Size, viewport: Size) -> int:
        """Get the optimal content width by arranging children.

        Args:
            widget: The container widget.
            container: The container size.
            viewport: The viewport size.

        Returns:
            Width of the content.
        """
        return widget.scrollable_content_region.width

    def get_content_height(
        self, widget: Widget, container: Size, viewport: Size, width: int
    ) -> int:
        """Get the content height.

        Args:
            widget: The container widget.
            container: The container size.
            viewport: The viewport.
            width: The content width.

        Returns:
            Content height (in lines).
        """
        if widget._nodes:
            arrangement = widget.arrange(Size(width, 0))
            height = arrangement.total_region.height
        else:
            height = 0
        return height



"""
Descriptors to define properties on your widget, screen, or App.

"""

from __future__ import annotations

from inspect import isclass
from typing import TYPE_CHECKING, Callable, Generic, TypeVar, overload

from textual._context import NoActiveAppError, active_app
from textual.css.query import NoMatches, QueryType, WrongType
from textual.widget import Widget

if TYPE_CHECKING:
    from textual.app import App
    from textual.dom import DOMNode
    from textual.message_pump import MessagePump


AppType = TypeVar("AppType", bound="App")


class app(Generic[AppType]):
    """Create a property to return the active app.

    All widgets have a default `app` property which returns an App instance.
    Type checkers will complain if you try to access attributes defined on your App class, which aren't
    present in the base class. To keep the type checker happy you can add this property to get your
    specific App subclass.

    Example:
        ```python
        class MyWidget(Widget):
            app = getters.app(MyApp)
        ```

    Args:
        app_type: The App subclass, or a callable which returns an App subclass.
    """

    def __init__(self, app_type: type[AppType] | Callable[[], type[AppType]]) -> None:
        self._app_type = app_type if isclass(app_type) else app_type()

    def __get__(self, obj: MessagePump, obj_type: type[MessagePump]) -> AppType:
        try:
            app = active_app.get()
        except LookupError:
            from textual.app import App

            node: MessagePump | None = obj
            while not isinstance(node, App):
                if node is None:
                    raise NoActiveAppError()
                node = node._parent
            app = node

        assert isinstance(app, self._app_type)
        return app


class query_one(Generic[QueryType]):
    """Create a query one property.

    A query one property calls [Widget.query_one][textual.dom.DOMNode.query_one] when accessed, and returns
    a widget. If the widget doesn't exist, then the property will raise the same exceptions as `Widget.query_one`.


    Example:
        ```python
        from textual import getters

        class MyScreen(screen):

            # Note this is at the class level
            output_log = getters.query_one("#output", RichLog)

            def compose(self) -> ComposeResult:
                with containers.Vertical():
                    yield RichLog(id="output")

            def on_mount(self) -> None:
                self.output_log.write("Screen started")
                # Equivalent to the following line:
                # self.query_one("#output", RichLog).write("Screen started")
        ```

    Args:
        selector: A TCSS selector, e.g. "#mywidget". Or a widget type, i.e. `Input`.
        expect_type: The type of the expected widget, e.g. `Input`, if the first argument is a selector.

    """

    selector: str
    expect_type: type["Widget"]

    @overload
    def __init__(self, selector: str) -> None:
        """

        Args:
            selector: A TCSS selector, e.g. "#mywidget"
        """

    @overload
    def __init__(self, selector: type[QueryType]) -> None: ...

    @overload
    def __init__(self, selector: str, expect_type: type[QueryType]) -> None: ...

    @overload
    def __init__(
        self, selector: type[QueryType], expect_type: type[QueryType]
    ) -> None: ...

    def __init__(
        self,
        selector: str | type[QueryType],
        expect_type: type[QueryType] | None = None,
    ) -> None:
        if expect_type is None:
            from textual.widget import Widget

            self.expect_type = Widget
        else:
            self.expect_type = expect_type
        if isinstance(selector, str):
            self.selector = selector
        else:
            self.selector = selector.__name__
            self.expect_type = selector

    @overload
    def __get__(
        self: "query_one[QueryType]", obj: DOMNode, obj_type: type[DOMNode]
    ) -> QueryType: ...

    @overload
    def __get__(
        self: "query_one[QueryType]", obj: None, obj_type: type[DOMNode]
    ) -> "query_one[QueryType]": ...

    def __get__(
        self: "query_one[QueryType]", obj: DOMNode | None, obj_type: type[DOMNode]
    ) -> QueryType | Widget | "query_one":
        """Get the widget matching the selector and/or type."""
        if obj is None:
            return self
        query_node = obj.query_one(self.selector, self.expect_type)
        return query_node


class child_by_id(Generic[QueryType]):
    """Create a child_by_id property, which returns the child with the given ID.

    This is similar using [query_one][textual.getters.query_one] with an id selector, except that
    only the immediate children are considered. It is also more efficient as it doesn't need to search the DOM.


    Example:
        ```python
        from textual import getters

        class MyScreen(screen):

            # Note this is at the class level
            output_log = getters.child_by_id("output", RichLog)

            def compose(self) -> ComposeResult:
                yield RichLog(id="output")

            def on_mount(self) -> None:
                self.output_log.write("Screen started")
        ```

    Args:
        child_id: The `id` of the widget to get (not a selector).
        expect_type: The type of the expected widget, e.g. `Input`.

    """

    child_id: str
    expect_type: type[Widget]

    @overload
    def __init__(self, child_id: str) -> None: ...

    @overload
    def __init__(self, child_id: str, expect_type: type[QueryType]) -> None: ...

    def __init__(
        self,
        child_id: str,
        expect_type: type[QueryType] | None = None,
    ) -> None:
        if expect_type is None:
            self.expect_type = Widget
        else:
            self.expect_type = expect_type
        self.child_id = child_id

    @overload
    def __get__(
        self: "child_by_id[QueryType]", obj: DOMNode, obj_type: type[DOMNode]
    ) -> QueryType: ...

    @overload
    def __get__(
        self: "child_by_id[QueryType]", obj: None, obj_type: type[DOMNode]
    ) -> "child_by_id[QueryType]": ...

    def __get__(
        self: "child_by_id[QueryType]", obj: DOMNode | None, obj_type: type[DOMNode]
    ) -> QueryType | Widget | "child_by_id":
        """Get the widget matching the selector and/or type."""
        if obj is None:
            return self
        child = obj._get_dom_base()._nodes._get_by_id(self.child_id)
        if child is None:
            raise NoMatches(f"No child found with id={self.child_id!r}")
        if not isinstance(child, self.expect_type):
            raise WrongType(
                f"Child with id={self.child_id!r} is the wrong type; expected type {self.expect_type.__name__!r}, found {child}"
            )
        return child



from __future__ import annotations

from rich.style import Style as RichStyle

from textual.color import Color
from textual.content import Style
from textual.css.styles import RulesMap
from textual.strip import Strip
from textual.visual import RenderOptions, Visual


class Blank(Visual):
    """Draw solid background color."""

    def __init__(self, color: Color | str = "transparent") -> None:
        self._rich_style = RichStyle.from_color(bgcolor=Color.parse(color).rich_color)

    def visualize(self) -> Blank:
        return self

    def get_optimal_width(self, rules: RulesMap, container_width: int) -> int:
        return container_width

    def get_height(self, rules: RulesMap, width: int) -> int:
        return 1

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        """Render the Visual into an iterable of strips. Part of the Visual protocol.

        Args:
            width: Width of desired render.
            height: Height of desired render or `None` for any height.
            style: The base style to render on top of.
            options: Additional render options.

        Returns:
            An list of Strips.
        """
        line_count = 1 if height is None else height
        return [Strip.blank(width, self._rich_style)] * line_count



from __future__ import annotations

from textual import events, on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Horizontal, ItemGrid, Vertical, VerticalScroll
from textual.demo._project_data import PROJECTS, ProjectInfo
from textual.demo._project_stars import STARS
from textual.demo.page import PageScreen
from textual.widgets import Footer, Label, Link, Markdown, Static

PROJECTS_MD = """\
# Projects

There are many amazing Open Source Textual apps available for download.
And many more still in development.

See below for a small selection!
"""


class Project(Vertical, can_focus=True, can_focus_children=False):
    """Display project information and open repo links."""

    ALLOW_MAXIMIZE = True
    DEFAULT_CSS = """
    Project {
        width: 1fr;
        height: auto;
        padding: 0 1;
        border: tall transparent;
        &:ansi {
            border: blank;
            &:focus {
                background: $panel;
            }
        }
        box-sizing: border-box;
        &:focus {
            border: tall $text-primary;
            background: $primary 20%;
            &.link {
                color: red !important;
            }
        }
        #title { text-style: bold; width: 1fr; }
        #author { text-style: italic; }
        .stars {
            color: $text-accent;
            text-align: right;
            text-style: bold;
            width: auto;
        }
        .header { height: 1; }
        .link {
            color: $text-accent;
            text-style: underline;
        }
        .description { color: $text-muted; }
        &.-hover { opacity: 1; }
    }
    """

    BINDINGS = [
        Binding(
            "enter",
            "open_repository",
            "open repo",
            tooltip="Open the GitHub repository in your browser",
        )
    ]

    def __init__(self, project_info: ProjectInfo) -> None:
        self.project_info = project_info
        super().__init__()

    def compose(self) -> ComposeResult:
        info = self.project_info
        with Horizontal(classes="header"):
            yield Label(info.title, id="title")
            yield Label(f"★ {STARS[info.title]}", classes="stars")
        yield Label(info.author, id="author")
        yield Link(info.url, tooltip="Click to open project repository")
        yield Static(info.description, classes="description")

    @on(events.Enter)
    @on(events.Leave)
    def on_enter(self, event: events.Enter):
        event.stop()
        self.set_class(self.is_mouse_over, "-hover")

    def action_open_repository(self) -> None:
        self.app.open_url(self.project_info.url)


class ProjectsScreen(PageScreen):
    AUTO_FOCUS = None
    CSS = """
    ProjectsScreen {
        align-horizontal: center;
        ItemGrid {
            margin: 2 4;
            padding: 1 2;
            background: $boost;
            width: 1fr;
            height: auto;
            grid-gutter: 1 1;
            grid-rows: auto;
            keyline:thin $foreground 30%;
            &:ansi {
                keyline: thin $border-blurred;
            }
        }
        Markdown { margin: 0; padding: 0 2; max-width: 100; background: transparent; }
    }
    """

    def compose(self) -> ComposeResult:
        with VerticalScroll() as container:
            container.can_focus = False
            with Center():
                yield Markdown(PROJECTS_MD)
            with ItemGrid(min_column_width=40):
                for project in PROJECTS:
                    yield Project(project)
        yield Footer()


if __name__ == "__main__":
    from textual.app import App

    class GameApp(App):
        def get_default_screen(self) -> Screen:
            return ProjectsScreen()

    app = GameApp()
    app.run()



from textual.geometry import Region


def get_auto_scroll_regions(
    widget_region: Region, auto_scroll_lines: int
) -> tuple[Region, Region]:
    """Get non-overlapping regions which should auto scroll when selecting.

    Args:
        widget_region: The region occupied by the widget.
        auto_scroll_lines: Number of lines in auto scroll regions.

    Returns:
        A pair of regions. The first for the region to scroll up, the second for the region to scroll down.
    """
    x, y, width, height = widget_region

    # Divide the region in to two, non overlapping regions
    top_half, bottom_half = widget_region.split_horizontal(height // 2)

    # Get a region at the top with the desired dimensions
    up_region = Region(x, y, width, auto_scroll_lines)
    # Ensure it is no larger than the top half
    up_region = top_half.intersection(up_region)

    # Repeat for the bottom half
    down_region = Region(x, y + height - auto_scroll_lines, width, auto_scroll_lines)
    down_region = bottom_half.intersection(down_region)

    return up_region, down_region



"""
`ScrollView` is a base class for [Line API](/guide/widgets#line-api) widgets.
"""

from __future__ import annotations

from rich.console import RenderableType

from textual._animator import EasingFunction
from textual._types import AnimationLevel, CallbackType
from textual.containers import ScrollableContainer
from textual.geometry import Region, Size


class ScrollView(ScrollableContainer):
    """
    A base class for a Widget that handles its own scrolling (i.e. doesn't rely
    on the compositor to render children).

    !!! note

        This is the typically wrong class for making something scrollable. If you want to make something scroll, set its
        `overflow` style to auto or scroll. Or use one of the pre-defined scrolling containers such as [VerticalScroll][textual.containers.VerticalScroll].
    """

    ALLOW_MAXIMIZE = True

    DEFAULT_CSS = """
    ScrollView {
        overflow-y: auto;
        overflow-x: auto;
    }
    """

    @property
    def is_scrollable(self) -> bool:
        """Always scrollable."""
        return True

    @property
    def is_container(self) -> bool:
        """Since a ScrollView should be a line-api widget, it won't have children,
        and therefore isn't a container."""
        return False

    def watch_scroll_x(self, old_value: float, new_value: float) -> None:
        if self.show_horizontal_scrollbar:
            self.horizontal_scrollbar.position = new_value
        if round(old_value) != round(new_value):
            self.refresh(self.size.region)

    def watch_scroll_y(self, old_value: float, new_value: float) -> None:
        if self.show_vertical_scrollbar:
            self.vertical_scrollbar.position = new_value
        if round(old_value) != round(new_value):
            self.refresh(self.size.region)

    def on_mount(self):
        self._refresh_scrollbars()

    def get_content_width(self, container: Size, viewport: Size) -> int:
        """Gets the width of the content area.

        Args:
            container: Size of the container (immediate parent) widget.
            viewport: Size of the viewport.

        Returns:
            The optimal width of the content.
        """
        return self.virtual_size.width

    def get_content_height(self, container: Size, viewport: Size, width: int) -> int:
        """Gets the height (number of lines) in the content area.

        Args:
            container: Size of the container (immediate parent) widget.
            viewport: Size of the viewport.
            width: Width of renderable.

        Returns:
            The height of the content.
        """
        return self.virtual_size.height

    def _size_updated(
        self, size: Size, virtual_size: Size, container_size: Size, layout: bool = True
    ) -> bool:
        """Called when size is updated.

        Args:
            size: New size.
            virtual_size: New virtual size.
            container_size: New container size.
            layout: Perform layout if required.

        Returns:
            True if a resize event should be sent, otherwise False.
        """
        if size_changed := self._size != size:
            self._set_dirty()
        if (
            size_changed
            or virtual_size != self.virtual_size
            or container_size != self.container_size
        ):
            self._scrollbar_changes.clear()
            self._size = size
            virtual_size = self.virtual_size
            self._container_size = size - self.styles.gutter.totals
            self._scroll_update(virtual_size)

        return size_changed or self._container_size != container_size

    def render(self) -> RenderableType:
        """Render the scrollable region (if `render_lines` is not implemented).

        Returns:
            Renderable object.
        """
        from rich.panel import Panel

        return Panel(f"{self.scroll_offset} {self.show_vertical_scrollbar}")

    # Custom scroll to which doesn't require call_after_refresh
    def scroll_to(
        self,
        x: float | None = None,
        y: float | None = None,
        *,
        animate: bool = True,
        speed: float | None = None,
        duration: float | None = None,
        easing: EasingFunction | str | None = None,
        force: bool = False,
        on_complete: CallbackType | None = None,
        level: AnimationLevel = "basic",
        immediate: bool = False,
    ) -> None:
        """Scroll to a given (absolute) coordinate, optionally animating.

        Args:
            x: X coordinate (column) to scroll to, or `None` for no change.
            y: Y coordinate (row) to scroll to, or `None` for no change.
            animate: Animate to new scroll position.
            speed: Speed of scroll if `animate` is `True`; or `None` to use `duration`.
            duration: Duration of animation, if `animate` is `True` and `speed` is `None`.
            easing: An easing method for the scrolling animation.
            force: Force scrolling even when prohibited by overflow styling.
            on_complete: A callable to invoke when the animation is finished.
            level: Minimum level required for the animation to take place (inclusive).
            immediate: If `False` the scroll will be deferred until after a screen refresh,
                set to `True` to scroll immediately.
        """

        self._scroll_to(
            x,
            y,
            animate=animate,
            speed=speed,
            duration=duration,
            easing=easing,
            force=force,
            on_complete=on_complete,
            level=level,
        )

    def refresh_line(self, y: int) -> None:
        """Refresh a single line.

        Args:
            y: Coordinate of line.
        """
        self.refresh(
            Region(
                0,
                y - self.scroll_offset.y,
                max(self.virtual_size.width, self.size.width),
                1,
            )
        )

    def refresh_lines(self, y_start: int, line_count: int = 1) -> None:
        """Refresh one or more lines.

        Args:
            y_start: First line to refresh.
            line_count: Total number of lines to refresh.
        """
        refresh_region = Region(
            0,
            y_start - self.scroll_offset.y,
            max(self.virtual_size.width, self.size.width),
            line_count,
        )
        self.refresh(refresh_region)



from textual.widgets._select import (
    NULL,
    EmptySelectError,
    InvalidSelectValueError,
    NoSelection,
)

__all__ = ["EmptySelectError", "InvalidSelectValueError", "NoSelection", "NULL"]



"""
Functions for *walking* the DOM.

!!! note

    For most purposes you would be better off using [query][textual.dom.DOMNode.query], which uses these functions internally.
"""

from __future__ import annotations

from collections import deque
from operator import attrgetter
from typing import TYPE_CHECKING, Iterable, Iterator, TypeVar, overload

from textual.geometry import Shape

if TYPE_CHECKING:
    from textual.dom import DOMNode
    from textual.widget import Widget

    WalkType = TypeVar("WalkType", bound=DOMNode)


if TYPE_CHECKING:

    @overload
    def walk_depth_first(
        root: DOMNode,
        *,
        with_root: bool = True,
    ) -> Iterable[DOMNode]: ...

    @overload
    def walk_depth_first(
        root: WalkType,
        filter_type: type[WalkType],
        *,
        with_root: bool = True,
    ) -> Iterable[WalkType]: ...


def walk_depth_first(
    root: DOMNode,
    filter_type: type[WalkType] | None = None,
    *,
    with_root: bool = True,
) -> Iterable[DOMNode] | Iterable[WalkType]:
    """Walk the tree depth first (parents first).

    !!! note

        Avoid changing the DOM (mounting, removing etc.) while iterating with this function.
        Consider [walk_children][textual.dom.DOMNode.walk_children] which doesn't have this limitation.

    Args:
        root: The root note (starting point).
        filter_type: Optional DOMNode subclass to filter by, or `None` for no filter.
        with_root: Include the root in the walk.

    Returns:
        An iterable of DOMNodes, or the type specified in `filter_type`.
    """
    stack: list[Iterator[DOMNode]] = [iter(root.children)]
    pop = stack.pop
    push = stack.append

    if filter_type is None:
        if with_root:
            yield root
        while stack:
            if (node := next(stack[-1], None)) is None:
                pop()
            else:
                yield node
                if children := node._nodes:
                    push(iter(children))
    else:
        if with_root and isinstance(root, filter_type):
            yield root
        while stack:
            if (node := next(stack[-1], None)) is None:
                pop()
            else:
                if isinstance(node, filter_type):
                    yield node
                if children := node._nodes:
                    push(iter(children))


if TYPE_CHECKING:

    @overload
    def walk_breadth_first(
        root: DOMNode,
        *,
        with_root: bool = True,
    ) -> Iterable[DOMNode]: ...

    @overload
    def walk_breadth_first(
        root: WalkType,
        filter_type: type[WalkType],
        *,
        with_root: bool = True,
    ) -> Iterable[WalkType]: ...


def walk_breadth_first(
    root: DOMNode,
    filter_type: type[WalkType] | None = None,
    *,
    with_root: bool = True,
) -> Iterable[DOMNode] | Iterable[WalkType]:
    """Walk the tree breadth first (children first).

    !!! note

        Avoid changing the DOM (mounting, removing etc.) while iterating with this function.
        Consider [walk_children][textual.dom.DOMNode.walk_children] which doesn't have this limitation.

    Args:
        root: The root note (starting point).
        filter_type: Optional DOMNode subclass to filter by, or `None` for no filter.
        with_root: Include the root in the walk.

    Returns:
        An iterable of DOMNodes, or the type specified in `filter_type`.
    """
    from textual.dom import DOMNode

    queue: deque[DOMNode] = deque()
    popleft = queue.popleft
    extend = queue.extend
    check_type = filter_type or DOMNode

    if with_root and isinstance(root, check_type):
        yield root
    extend(root.children)
    while queue:
        node = popleft()
        if isinstance(node, check_type):
            yield node
        extend(node._nodes)


def walk_breadth_search_id(
    root: DOMNode, node_id: str, *, with_root: bool = True
) -> DOMNode | None:
    """Special case to walk breadth first searching for a node with a given id.

    This is more efficient than [walk_breadth_first][textual.walk.walk_breadth_first] for this special case, as it can use an index.

    Args:
        root: The root node (starting point).
        node_id: Node id to search for.
        with_root: Consider the root node? If the root has the node id, then return it.

    Returns:
        A DOMNode if a node was found, otherwise `None`.
    """

    if with_root and root.id == node_id:
        return root

    queue: deque[DOMNode] = deque()
    queue.append(root)

    while queue:
        node = queue.popleft()
        if (found_node := node._nodes._get_by_id(node_id)) is not None:
            return found_node
        queue.extend(node._nodes)
    return None


def walk_selectable_widgets(
    root: DOMNode, bounds: Shape, bounded: set[DOMNode]
) -> Iterable[Widget]:
    """Walk the tree depth first in select order (top to bottom, then left to right).

    Args:
        root: The root note (starting point).
        bounds: A Shape object that defines the selection bounds.
        bounded: Container widgets that require a bounds check.

    Returns:
        An iterable of DOMNodes.
    """
    stack: list[Iterator[Widget]] = [iter(root.children)]
    pop = stack.pop
    push = stack.append

    get_selection_order = attrgetter("_selection_order")

    def get_children(node: DOMNode) -> list[Widget]:
        """Get children, sorted in selection order, and potentially filtered by selection bounds.

        Args:
            node: A root node.

        Returns:
            A list of child widgets.
        """
        children = sorted(
            node.displayed_and_visible_children,
            key=get_selection_order,
        )
        if node in bounded:
            children = [child for child in children if bounds.overlaps(child.region)]
        return children

    children = get_children(root)

    while stack:
        if (node := next(stack[-1], None)) is None:
            pop()
        elif node.allow_select:
            yield node
            if children := get_children(node):
                push(iter(children))



"""
Timer context manager, only used in debug.
"""

import contextlib
from time import perf_counter
from typing import Generator

from textual import log


@contextlib.contextmanager
def timer(subject: str = "time", threshold: float = 0) -> Generator[None, None, None]:
    """print the elapsed time. (only used in debugging).

    Args:
        subject: Text shown in log.
        threshold: Time in second after which the log is written.

    """
    start = perf_counter()
    yield
    elapsed = perf_counter() - start
    if elapsed >= threshold:
        elapsed_ms = elapsed * 1000
        log(f"{subject} elapsed {elapsed_ms:.4f}ms")
